@@ -1,5 +1,5 @@
 """
-Copyright 2019-2020 by J. Christopher Wagner (jwag). All rights reserved.
+Copyright 2019-2024 by J. Christopher Wagner (jwag). All rights reserved.
 :license: MIT, see LICENSE for more details.
 
 
@@ -11,21 +11,20 @@ BE AWARE: Once any version of this is shipped no changes can be made - instead
 a new version needs to be created.
 """
 
-import datetime
+from typing import cast
 from sqlalchemy import (
     Boolean,
     DateTime,
     Column,
     Integer,
     String,
-    UnicodeText,
     ForeignKey,
 )
 from sqlalchemy.ext.declarative import declared_attr
-from sqlalchemy.orm import relationship
+from sqlalchemy.ext.mutable import MutableList
 from sqlalchemy.sql import func
 
-from flask_security import RoleMixin, UserMixin
+from flask_security import AsaList, RoleMixin, UserMixin, naive_utcnow
 
 
 class FsModels:
@@ -40,16 +39,29 @@ class FsModels:
     fs_model_version = 1
     user_table_name = "user"
     role_table_name = "role"
+    webauthn_table_name = "webauthn"
 
     @classmethod
-    def set_db_info(cls, appdb, user_table_name="user", role_table_name="role"):
+    def set_db_info(
+        cls,
+        appdb,
+        user_table_name="user",
+        role_table_name="role",
+        webauthn_table_name="webauthn",
+    ):
         """Initialize Model.
         This needs to be called after the DB object has been created
-        (e.g. db = Sqlalchemy())
+        (e.g. db = Sqlalchemy()).
+
+        .. note::
+            This should only be used if you are utilizing the fsqla data
+            models. With your own models you would need similar but slightly
+            difficult code.
         """
         cls.db = appdb
         cls.user_table_name = user_table_name
         cls.role_table_name = role_table_name
+        cls.webauthn_table_name = webauthn_table_name
         cls.roles_users = appdb.Table(
             "roles_users",
             Column("user_id", Integer(), ForeignKey(f"{cls.user_table_name}.id")),
@@ -62,12 +74,14 @@ class FsRoleMixin(RoleMixin):
     name = Column(String(80), unique=True, nullable=False)
     description = Column(String(255))
     # A comma separated list of strings
-    permissions = Column(UnicodeText, nullable=True)
+    permissions = Column(
+        MutableList.as_mutable(AsaList()), nullable=True  # type: ignore
+    )
     update_datetime = Column(
-        DateTime,
+        type_=DateTime,
         nullable=False,
         server_default=func.now(),
-        onupdate=datetime.datetime.utcnow,
+        onupdate=naive_utcnow,
     )
 
 
@@ -80,7 +94,7 @@ class FsUserMixin(UserMixin):
     # Username is important since shouldn't expose email to other users in most cases.
     username = Column(String(255))
     password = Column(String(255), nullable=False)
-    active = Column(Boolean(), nullable=False)
+    active = cast(bool, Column(Boolean(), nullable=False))
 
     # Flask-Security user identifier
     fs_uniquifier = Column(String(64), unique=True, nullable=False)
@@ -106,65 +120,15 @@ class FsUserMixin(UserMixin):
         return FsModels.db.relationship(
             "Role",
             secondary=FsModels.roles_users,
-            backref=FsModels.db.backref("users", lazy="dynamic"),
+            backref=FsModels.db.backref(
+                "users", lazy="dynamic", cascade_backrefs=False
+            ),
         )
 
-    create_datetime = Column(DateTime, nullable=False, server_default=func.now())
+    create_datetime = Column(type_=DateTime, nullable=False, server_default=func.now())
     update_datetime = Column(
-        DateTime,
+        type_=DateTime,
         nullable=False,
         server_default=func.now(),
-        onupdate=datetime.datetime.utcnow,
+        onupdate=naive_utcnow,
     )
-
-
-"""
-These are placeholders - not current used
-"""
-
-
-class FsOauth2ClientMixin:
-    """ Oauth2 client """
-
-    id = Column(String(64), primary_key=True)
-
-    @declared_attr
-    def user_id(cls):
-        return Column(
-            Integer, ForeignKey("user.id", ondelete="CASCADE"), nullable=False
-        )
-
-    @declared_attr
-    def user(cls):
-        return relationship("User")
-
-    grant_type = Column(String(32), nullable=False)
-    scopes = Column(UnicodeText(), default="")
-    response_type = Column(UnicodeText, nullable=False, default="")
-    redirect_uris = Column(UnicodeText())
-
-
-class FsTokenMixin:
-    """ (Bearer) Tokens that have been given out """
-
-    id = Column(Integer, primary_key=True)
-
-    @declared_attr
-    def client_id(cls):
-        return Column(
-            Integer, ForeignKey("oauth2_client.id", ondelete="CASCADE"), nullable=False
-        )
-
-    # client = relationship("fs_oauth2_client")
-    @declared_attr
-    def user_id(cls):
-        return Column(
-            Integer, ForeignKey("user.id", ondelete="CASCADE"), nullable=False
-        )
-
-    scopes = Column(UnicodeText(), default="")
-    revoked = Column(Boolean(), nullable=False, default=False)
-    access_token = Column(String(100), unique=True, nullable=False)
-    refresh_token = Column(String(100), unique=True)
-    issued_at = Column(DateTime, nullable=False, server_default=func.now())
-    expires_at = Column(DateTime())
